int get_archive_bit(char *);
char **expargs (int *, char **);
char ** check_file_option (char *this_item,int *pargc,char **av);
char ** add_arg(char *this_arg,int *pargc,char **new_argv) ;
static char ** grow_argv (char **new_argv, int new_argc);

/* dosargs.c - was: 
 * tcexparg.c - Unix-style command line wildcards for Turbo C 2.0
 *
 * now also handles @file in typical dos style by including the
 * contents of the file into the arg list, and does simple quote
 * handling by grouping everything from the 1st ' or " to a matching
 * character into one argument.
 */
/*
   This file is in the public domain.

   Syntax for wildcards:
   *		Matches zero or more of any character (except a '.' at
		the beginning of a name).
   ?		Matches any single character.
   [r3z]	Matches 'r', '3', or 'z'.
   [a-d]	Matches a single character in the range 'a' through 'd'.
   [!a-d]	Matches any single character except a character in the
		range 'a' through 'd'.

   The period between the filename root and its extension need not be
   given explicitly.  Thus, the pattern `a*e' will match 'abacus.exe'
   and 'axyz.e' as well as 'apple'.  Comparisons are not case sensitive.

   Authors:
   The expargs code is a modification of wildcard expansion code
   written for Turbo C 1.0 by
   Richard Hargrove
   Texas Instruments, Inc.
   P.O. Box 869305, m/s 8473
   Plano, Texas 75086
   214/575-4128
   and posted to USENET in September, 1987.

   The wild_match code was written by Rich Salz, rsalz@bbn.com,
   posted to net.sources in November, 1986.

   The code connecting the two is by Mike Slomin, bellcore!lcuxa!mike2,
   posted to comp.sys.ibm.pc in November, 1988.

   Major performance enhancements and bug fixes, and source cleanup,
   by David MacKenzie, djm@gnu.ai.mit.edu.

   Rearranged for @file inclusions by Les Mikesell, March 1994.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <dos.h>
#include <dir.h>
#include "port.h"
#include "tar.h"

PTR ck_malloc (size_t size);

/* Number of new arguments to allocate space for at a time.  */
#define ARGS_INCREMENT 20

static char **grow_argv (char **new_argv, int new_argc);
static void fatal_error (const char *message);
int wild_match (char *string, char *pattern);
char *basename (char *path);
char **expargs (int *, char **);

  char path[MAXPATH + 1];
  char **new_argv;
  struct ffblk block;
  char *arg_base;
  int path_length;
  int matched;


/* args coming from command line already have dos-style quote
   processing done - just check for file inclusions
 */
char **
expargs (int *pargc, char **argv)
{
  int new_argc;
  int argind;
  int argcount = *pargc;
  new_argv = grow_argv ((char **)NULL, 0);
  new_argv[0] = basename(argv[0]);
  strlwr(new_argv[0]);
  tar = new_argv[0]; 
  *pargc = 1;
  for (argind = 1; argind < argcount; ++argind) {
	new_argv = check_file_option(argv[argind],pargc,new_argv);
  }
  return (new_argv); 
}

/* Split a line into individual options.
 * Keep enclosed item in one option preserving spaces. 
 * Leave single quotes for the next level to inhibit @file and wildcards.
 *   quotes can't span lines
 */
char **
parse_option(char *input,int *pargc,char **av)
{
	char *current,*end;
	if (av == 0 ) {
	  av = grow_argv ((char **)NULL, 0);
	}
	for (current=input;*current;) {
		while (isspace(*current)) current++;
		if (*current == '\0')
			break;
		end = current;
		switch (*current) {
		case  '"':
			end = ++current; /* strip doubles */
			while (*end && *end != '"') end++;
			break ;
		 case '\'' :
			end++; /* leave singles */
			while (*end && *end != '\'') end++;
			if (*end) end++;
			break ;
		default:
			while (*end++) {
				if (isspace(*end)) break;
			}
		}
		if (*end) {
			*end= '\0';
			av = check_file_option(current,pargc,av);
			current=end+1;
		} else {
			av = check_file_option(current,pargc,av);
			current = end;
		}
	}
	return (av);
}

	
/* here we look for @file inclusions, possibly recursing with the contents */
char **
check_file_option (char *this_item,int *pargc,char **av)
{
#define WORKSPACE 1024   /* probably should be bigger */ 
  char work[WORKSPACE];
  FILE *argfile;
  char *fargptr;
  int fargsize;
      if (*this_item == '@') {
	if ((argfile=fopen(++this_item,"r")) == NULL) {
	  sprintf(work,"Can't open argument file %s",this_item);
	  fatal_error (work);
	}
	while (fgets(work,WORKSPACE,argfile)) {
		if (strlen(work) >= WORKSPACE -1) { /* bail out */	
		  sprintf(work,"Argument file %s line too large",this_item);
		  fatal_error (work);
   		 }
		    fargptr=(char *)ck_malloc(strlen(work)+1); 
		    strcpy(fargptr,work);            /* give permanent home */
		av = parse_option(fargptr,pargc,av); /* note - will recurse */
	}
	fclose(argfile);
      } else {
	av = add_arg(this_item,pargc,av);
	} 
	return (av);
} 

/* single quotes surrounding item should be stripped and
 * the enclosed value should not be expanded
 *   Otherwise you can't use wildcards for extracting...
 */
char **
add_arg(char *this_arg,int *pargc,char **new_argv) 
{
  char *path_base;
  int new_argc = *pargc;
      matched = 0;
      if (*this_arg == '\'')  {  /* check for single quotes */
	this_arg++;
	if ((path_base = strchr(this_arg+1,'\'')) != NULL) {
		*path_base='\0';  /* zap off both ends */
	}
      } else if (strpbrk (this_arg, "?*[") != NULL) {
	  strncpy (path, this_arg, MAXPATH - 3);
	  path_base = basename (path);
	  strcpy (path_base, "*.*");
	  arg_base = this_arg + (path_base - path);

	  if (!findfirst (path, &block, FA_DIREC))
	    {
	      strlwr (path);
	      do
		{
		  /* Only match "." and ".." explicitly.  */
		  if (*block.ff_name == '.' && *arg_base != '.')
		    continue;
		  path_length = stpcpy (path_base, block.ff_name) - path + 1;
		  strlwr (path_base);
		  if (wild_match (path, this_arg))
		    {
		      matched = 1;
		      new_argv[new_argc] = (char *) malloc (path_length);
		      if (new_argv[new_argc] == NULL)
			fatal_error ("memory exhausted");
		      strcpy (new_argv[new_argc++], path);
		      new_argv = grow_argv (new_argv, new_argc);
		    }
	      } while (!findnext (&block));
	    }
	}
      if (matched == 0)
	new_argv[new_argc++] = this_arg;
      new_argv = grow_argv (new_argv, new_argc);
  *pargc = new_argc;
  new_argv[new_argc] = NULL;
  return (new_argv);
}

/* Return a pointer to the last element of PATH.  */

char *
basename (char *path)
{
  char *tail;

  for (tail = path; *path; ++path)
    if (*path == ':' || *path == '/')
      tail = path + 1;
  return tail;
}

static char **
grow_argv (char **new_argv, int new_argc)
{
  if (new_argv == 0) 
    {
      new_argv = (char **) malloc
	(sizeof (char *) * (new_argc + ARGS_INCREMENT));
    } else {
  if (new_argc % ARGS_INCREMENT == 0)
    {
      new_argv = (char **) realloc
	(new_argv, sizeof (char *) * (new_argc + ARGS_INCREMENT));
    }
  if (new_argv == NULL)
	fatal_error ("memory exhausted");
	}
  return new_argv;
}

static void
fatal_error (const char *message)
{
  putc ('\n', stderr);
  if (tar && *tar)
    {
      fputs (tar, stderr);
      fputs (": ", stderr);
    }
  fputs (message, stderr);
  putc ('\n', stderr);
  exit (1);
}

/* Shell-style pattern matching for ?, \, [], and * characters.
   I'm putting this replacement in the public domain.

   Written by Rich $alz, mirror!rs, Wed Nov 26 19:03:17 EST 1986.  */

/* The character that inverts a character class; '!' or '^'.  */
#define INVERT '!'

static int star (char *string, char *pattern);

/* Return nonzero if `string' matches Unix-style wildcard pattern
   `pattern'; zero if not.  */

int
wild_match (char *string, char *pattern)
{
  int prev;		/* Previous character in character class.  */
  int matched;		/* If 1, character class has been matched.  */
  int reverse;		/* If 1, character class is inverted.  */

  for (; *pattern; string++, pattern++)
    switch (*pattern)
      {
      case '\\':
	/* Literal match with following character; fall through.  */
	pattern++;
      default:
	if (*string != *pattern)
	  return 0;
	continue;
      case '?':
	/* Match anything.  */
	if (*string == '\0')
	  return 0;
	continue;
      case '*':
	/* Trailing star matches everything.  */
	return *++pattern ? star (string, pattern) : 1;
      case '[':
	/* Check for inverse character class.  */
	reverse = pattern[1] == INVERT;
	if (reverse)
	  pattern++;
	for (prev = 256, matched = 0; *++pattern && *pattern != ']';
	     prev = *pattern)
	  if (*pattern == '-'
	      ? *string <= *++pattern && *string >= prev
	      : *string == *pattern)
	    matched = 1;
	if (matched == reverse)
	  return 0;
	continue;
      }

  return *string == '\0';
}

static int
star (char *string, char *pattern)
{
  while (wild_match (string, pattern) == 0)
    if (*++string == '\0')
      return 0;
  return 1;
}
/* these routines make the -C option handle dos switching */
#ifdef chdir
#undef chdir
#endif

static char *dospath;
static int dosdrive;
int dos_error_handler(int); 

/* this is called at start-up 
 * save current drive and working directory
 * also install hardware error handler since it is
 * normal for DOS to get errors on files that are
 * open and we don't want to get the abort/retry/fail message
 */

void
savecwd(void)
{
    char buf[PATH_MAX+1];
    dosdrive = getdisk();
    getcwd(buf,PATH_MAX);
    dospath=(char *)ck_malloc(strlen(buf)+1);
    strcpy(dospath,buf);
    harderr(dos_error_handler);
    return;
}

void
resetcwd(void)
{
    setdisk(dosdrive);
    chdir(dospath);
    return;
}

/* allow -C to work with drive specifications under dos */
/* NOTE: non-absolute dirs are relative to starting drive/dir */
int
dos_chdir(char *path)
{
   int ret;
   resetcwd();  /* back to starting point */
   ret = chdir(path);
   if (ret) 
	return (ret);
   if (path[1] == ':' ) {
	ret = setdisk(*path - 'A');
   }
   return (ret);
}

/* we expect this to run unattended - failures should be reported */
dos_error_handler(int dos_err)
{
       hardresume(_HARDERR_FAIL);
}

/* clear the archive bit */
void
reset_archive(char *filename)
{
	unsigned attrib;
	if (_dos_getfileattr(filename,&attrib) != 0) {
	 	fprintf(stderr,"Can't get attributes of %s\n",filename);
		return;
	}
	if ((attrib & FA_ARCH) == 0 ) {
		return;
	}
	attrib &= ~FA_ARCH;
	if (_dos_setfileattr(filename,attrib) != 0) {
	 	fprintf(stderr,"Can't set attributes of %s\n",filename);
	}
	return;
}

/* return non-zero if file is a directory or has the archive bit clear.
 *  actually I don't think directories can ever have the archive bit clear
 *   but it doesn't hurt to be safe
 */
int
get_archive_bit(char *filename)
{
	unsigned attrib;
	if (_dos_getfileattr(filename,&attrib) != 0) {
	 	fprintf(stderr,"Can't get attributes of %s\n",filename);
		return (1);
	}
	if (attrib & FA_DIREC) {
		return (1);
	}
	if (attrib & FA_ARCH) {
		return (1);
	}
	return (0);
}

/* twiddle HIDDEN and SYSTEM bits into group and other execute perms
 * HIDDEN = st_mode |= 0010, SYSTEM = |= 0001 
 *  this will match GTAK settings
 * then add the gid mapping used by AT&T/NCR's lanman server
 *   also set r/w and execute bits on directories
 */

static struct {
  int d_gid;     /* At&t stargroup lanman server group id */
  char *d_gname; /* and matching name */
  int d_attrib;  /* dos bits */
 } dosgrp[] = {
 {99,"DOS----",0},
 {98,"DOS-a--",FA_ARCH},
 {97,"DOS--s-",FA_SYSTEM},
 {96,"DOS---h",FA_HIDDEN},
 {95,"DOS-as-",FA_ARCH | FA_SYSTEM},
 {94,"DOS-a-h",FA_ARCH |FA_HIDDEN},
 {93,"DOS--sh",FA_SYSTEM|FA_HIDDEN},
 {92,"DOS-ash",FA_ARCH |FA_SYSTEM |FA_HIDDEN},
 {-1,"",-1}
};

/* get dos file attributes into stat struct
 * force rwx modes on directories
 * if archive bit is clear on regular files, return 0
 */

int
dos_stat(char *filename, struct stat *st) 
{
	unsigned attrib;
	int i = 0;
 	int err, isdir;
	if ((err=stat(filename,st)) != 0) 
		return err;
	if ((err = _dos_getfileattr(filename,&attrib)) != 0) {
		return err;
	}
	if ((isdir = attrib & FA_DIREC) != 0) {
		st->st_mode |= S_IREAD|S_IWRITE|S_IEXEC ;
	}
	if (attrib & FA_HIDDEN) {
		st->st_mode |= 0010 ;
	}
	if (attrib & FA_SYSTEM) {
		st->st_mode |= 0001 ;
	}
	attrib &= FA_HIDDEN | FA_SYSTEM | FA_ARCH ;
	while (dosgrp[i].d_gid >= 0 ) {
		if (attrib == dosgrp[i].d_attrib) {
			st->st_gid = dosgrp[i].d_gid;
			break;
		}
		i++;
	}
        if (isdir) { /* a directory */
		return (isdir); 
	}
	return ((int) attrib & FA_ARCH);
}

/* this is called after a file is written if f_reset_archive or
 * f_use_protection is set
 */
void
dos_fixmodes(char *filename, int tarmodes)
{
	unsigned was,new,check;
	new = 0;
	if (f_use_protection) { 
		if (tarmodes & 0010)
			new |= FA_HIDDEN;
		if (tarmodes & 0001)
			new |= FA_SYSTEM;
	}
	if (!(new || f_reset_archive) )
		return;  /* nothing to do */
	if (_dos_getfileattr(filename,&was) != 0) {
	 	fprintf(stderr,"Can't get attributes of %s\n",filename);
		return;
	}
	new |= was;
	if (f_reset_archive)
		new &= ~FA_ARCH;
	if (new != was) {
		if (_dos_setfileattr(filename,new) != 0) {
			fprintf(stderr,"Can't set attributes of %s\n",filename);
		}
	}
	return;
}

static char saveuname[TUNMLEN];

void
dos_finduname (char *uname,int uid)
{
  char *p, *p1;
  if (!*saveuname) { 
	if (f_uname && *f_uname) {
		strncpy (saveuname,f_uname,TUNMLEN);
	} else {
		p=f_envnames;
		do {
			if ((p1=strchr(p,':')) !=0) {
				*p1++ = '\0';
			}
			if ((p=getenv(p)) != 0 && *p != 0) {
				strncpy(saveuname,p,TUNMLEN);
				break;
			}
			p=p1;
		} while (p && *p);
	}
	if ( !*saveuname) strcpy(saveuname,"DOS");
  }
  strncpy (uname, saveuname, TUNMLEN);
}

/* not much we can do here */
int
dos_finduid (uname)
{
  return f_uid;
}

void
dos_findgname (char *gname, int gid)
{
  int i = 0;
  char *p = " ";
  if (f_gname)  {
	p = f_gname;
  } else {
	  while (dosgrp[i].d_gid >= 0 ) {
		if (gid == dosgrp[i].d_gid){
		   p=dosgrp[i].d_gname;
		   break;
	        }
		i++;
	  }
  }
	  (void) strncpy (gname, p, TGNMLEN);
}

/* We'll return the dos attributes encoded in the gid */
int
dos_findgid (char *gname)
{
	int i = 0;
	if (f_gid >= 0) return (f_gid);  /* command line override */
	  while (dosgrp[i].d_gname >= 0 ) {
		if (strcmp(gname,dosgrp[i].d_gname) == 0) {
		   return (dosgrp[i].d_gid);
		}
		i++;
	  }
  return 0;
}
void
do_rsh_restart(void)  {
	execl(f_prog_path,f_prog_path, "--wait", "--rshd", (char *) NULL);
}

